import pytest

from creevey.load_funcs.image import load_image_from_disk
from tests.conftest import SAMPLE_DATA_DIR


@pytest.fixture
def sample_image_square_rgb():
    filename = 'creevey_rgb.jpg'
    path = SAMPLE_DATA_DIR / filename
    image = load_image_from_disk(path)
    return image


@pytest.fixture
def sample_image_tall_grayscale():
    filename = 'creevey_gray.jpg'
    path = SAMPLE_DATA_DIR / filename
    image = load_image_from_disk(path)
    return image


@pytest.fixture
def sample_image_square_rgba():
    filename = 'creevey_rgba.png'
    path = SAMPLE_DATA_DIR / filename
    image = load_image_from_disk(path)
    return image



from collections import defaultdict

from creevey.ops import get_report_output_decorator, report_output


def test_report_output():
    log_dict = defaultdict(dict)
    result = report_output(
        func=lambda x: x + 1,
        func_input=1,
        key='plus1',
        inpath='fake',
        log_dict=log_dict,
    )
    assert result == 1
    assert log_dict['fake']['plus1'] == 2


def test_get_report_output_decorator():
    log_dict = defaultdict(dict)

    @get_report_output_decorator(key='plus1')
    def report_plus1(num):
        return num + 1

    result = report_plus1(1, inpath='fake', log_dict=log_dict)
    assert result == 1
    assert log_dict['fake']['plus1'] == 2



import cv2 as cv
import numpy as np

from creevey.ops.helpers.report import get_report_output_decorator
from creevey.ops.image.transforms import convert_to_grayscale


def calculate_mean_brightness(image: np.array) -> float:
    """
    Calculate mean image brightness

    Brightness is calculated by converting to grayscale if necessary and
    then taking the mean pixel value. Assumes image is grayscale, RGB,
    or RGBA.
    """
    return convert_to_grayscale(image).mean()


@get_report_output_decorator(key='mean_brightness')
def report_mean_brightness(image):
    return calculate_mean_brightness(image)


def calculate_dhash(image: np.array, sqrt_hash_size: int = 8) -> np.array:
    """
    Calculate difference hash of image.

    As a rule of thumb, with `sqrt_hash_size=8`, hashes from two images
    should typically have a Hamming distance less than 10 if and only if
    those images are "duplicates", with some robustness to sources of
    noise such as resizing and JPEG artifacts, where the Hamming
    distance between two hashes `a` and `b` is computed as
    `bin(a ^ b).count("1")`.

    Assumes image is grayscale, RGB, or RGBA.

    Source
    ------
    Adrian Rosebrock, "Building an Image Hashing Search Engine with
    VP-Trees and OpenCV", *PyImageSearch*,
    https://www.pyimagesearch.com/2019/08/26/building-an-image-hashing-search-engine-with-vp-trees-and-opencv/,
    accessed on 18 October 2019.

    Parameters
    ----------
    image
    sqrt_hash_size
        Side length of 2D array used to compute hash, so that hash will
        be up to `sqrt_hash_size`^2 bits long.
    """
    im_mod = convert_to_grayscale(image)
    im_mod = cv.resize(im_mod, (sqrt_hash_size + 1, sqrt_hash_size))
    diff = im_mod[:, 1:] > im_mod[:, :-1]
    return sum([2 ** i for (i, v) in enumerate(diff.flatten()) if v])


@get_report_output_decorator(key='dhash')
def report_dhash(image, sqrt_hash_size: int = 8):
    return calculate_dhash(image=image, sqrt_hash_size=sqrt_hash_size)



"""Code for reporting information calculating during processing"""
from functools import partial
from typing import Any, Callable, DefaultDict, Hashable

from creevey.constants import PathOrStr


def report_output(
    func_input: Any,
    func: Callable,
    inpath: PathOrStr,
    log_dict: DefaultDict[str, dict],
    key: Hashable,
) -> Any:
    """
    Add the output of a function to `log_dict[inpath][key]`.

    Return the input to that function in order to pass it along within a
    pipeline.

    Intended to be used within a `CustomReportingPipeline` to add the
    output of the function to the run report.

    Examples
    --------
    >> from functools import partial
    >>
    >> from creevey import CustomReportingPipeline
    >> from creevey.load_funcs.image import load_image_from_url
    >> from creevey.ops import report_output
    >> from creevey.ops.image import calculate_mean_brightness
    >> from creevey.path_funcs import join_outdir_filename_extension
    >> from creevey.write_funcs.image import write_image
    >>
    >> report_mean_brightness = partial(
    >>     report_output, func=calculate_mean_brightness, key='mean_brightness'
    >> )
    >>
    >> report_brightness_pipeline = CustomReportingPipeline(
    >>     load_func=load_image_from_url, ops=[report_mean_brightness], write_func=write_image
    >> )
    >>
    >> image_filenames = ['2RsJ8EQ', '2TqoToT', '2VocS58', '2scKPIp', '2TsO6Pc', '2SCv0q7']
    >> image_urls = [f'https://bit.ly/{filename}' for filename in image_filenames]
    >>
    >> keep_filename_png_in_cwd = partial(
    >>     join_outdir_filename_extension, outdir='.', extension='.png'
    >> )
    >> print(
    >>     report_brightness_pipeline.run(
    >>         inpaths=image_urls,
    >>         path_func=keep_filename_png_in_cwd,
    >>         n_jobs=1,
    >>         skip_existing=False,
    >>     )
    >> )
                                mean_brightness  ... time_finished
    https://bit.ly/2RsJ8EQ        78.570605  ...  1.571842e+09
    https://bit.ly/2SCv0q7       130.348113  ...  1.571842e+09
    https://bit.ly/2TqoToT        82.677745  ...  1.571842e+09
    https://bit.ly/2TsO6Pc       151.596546  ...  1.571842e+09
    https://bit.ly/2VocS58        72.072578  ...  1.571842e+09
    https://bit.ly/2scKPIp       117.491313  ...  1.571842e+09

    Parameters
    ----------
    func_input
        Input to `func`
    func
        Function whose output is to be reported
    inpath
        Input path associated with `func_input`
    log_dict
        Dictionary for storing function output (in
        `log_dict[inpath][key]`)
    key
        Dictionary key in which to store function output for each
        inpath

    Returns
    -------
    func_input

    Side effect
    -----------
    Assigns `func(func_input)` to `log_dict[inpath][key]`
    """
    result = func(func_input)
    log_dict[inpath][key] = result
    return func_input


def get_report_output_decorator(key: Hashable) -> Callable:
    """
    Get a decorator that modifies a function to add its output to
    `log_dict[inpath][key]` and return its input.

    Intended to be used to adapt a function for use within a
    `CustomReportingPipeline`.

    Examples
    --------
    >> from functools import partial
    >>
    >> from creevey import CustomReportingPipeline
    >> from creevey.load_funcs.image import load_image_from_url
    >> from creevey.ops import get_report_output_decorator
    >> from creevey.ops.image import calculate_mean_brightness
    >> from creevey.path_funcs import join_outdir_filename_extension
    >> from creevey.write_funcs.image import write_image
    >>
    >>
    >> @get_report_output_decorator(key='mean_brightness')
    >> def report_mean_brightness(image):
    >>     return calculate_mean_brightness(image)
    >>
    >>
    >> report_brightness_pipeline = CustomReportingPipeline(
    >>     load_func=load_image_from_url, ops=[report_mean_brightness], write_func=write_image
    >> )
    >>
    >> image_filenames = ['2RsJ8EQ', '2TqoToT', '2VocS58', '2scKPIp', '2TsO6Pc', '2SCv0q7']
    >> image_urls = [f'https://bit.ly/{filename}' for filename in image_filenames]
    >>
    >> keep_filename_png_in_cwd = partial(
    >>     join_outdir_filename_extension, outdir='.', extension='.png'
    >> )
    >> print(
    >>     report_brightness_pipeline.run(
    >>         inpaths=image_urls,
    >>         path_func=keep_filename_png_in_cwd,
    >>         n_jobs=1,
    >>         skip_existing=False,
    >>     )
    >> )
                                mean_brightness  ... time_finished
    https://bit.ly/2RsJ8EQ        78.570605  ...  1.571843e+09
    https://bit.ly/2SCv0q7       130.348113  ...  1.571843e+09
    https://bit.ly/2TqoToT        82.677745  ...  1.571843e+09
    https://bit.ly/2TsO6Pc       151.596546  ...  1.571843e+09
    https://bit.ly/2VocS58        72.072578  ...  1.571843e+09
    https://bit.ly/2scKPIp       117.491313  ...  1.571843e+09


    Parameters
    ----------
    key
        Dictionary key in which to store function output for each
        inpath
    """
    # noqa: D202
    def report_output_decorator(func):
        return partial(report_output, func=func, key=key)

    return report_output_decorator



from collections import defaultdict

import numpy as np

from creevey.ops.image.stats import (
    calculate_dhash,
    calculate_mean_brightness,
    report_dhash,
    report_mean_brightness,
)
from creevey.ops.image.transforms import resize
from tests.conftest import SAMPLE_DATA_DIR


class TestCalculateMeanBrightness:
    def test_calculate_mean_brightness_known_val(self):
        assert calculate_mean_brightness(np.array([[0, 0, 0], [1, 1, 1]])) == 0.5

    def test_calculate_mean_brightness_gray(self, sample_image_tall_grayscale):
        assert isinstance(calculate_mean_brightness(sample_image_tall_grayscale), float)

    def test_calculate_mean_brightness_rgb(self, sample_image_square_rgb):
        assert isinstance(calculate_mean_brightness(sample_image_square_rgb), float)

    def test_calculate_mean_brightness_rgba(self, sample_image_square_rgba):
        assert isinstance(calculate_mean_brightness(sample_image_square_rgba), float)


class TestReportMeanBrightness:
    def test_report_mean_brightness(self, sample_image_square_rgb):
        log_dict = defaultdict(dict)
        image_path = SAMPLE_DATA_DIR / 'creevey_rgb.jpg'
        result = report_mean_brightness(
            sample_image_square_rgb, log_dict=log_dict, inpath=image_path
        )
        assert log_dict[image_path]['mean_brightness'] == calculate_mean_brightness(
            sample_image_square_rgb
        )
        assert (result == sample_image_square_rgb).all()


class TestCalculateDHash:
    def test_calculate_dhash_rgb_hashlen_rgb(self, sample_image_square_rgb):
        hash = calculate_dhash(sample_image_square_rgb, sqrt_hash_size=8)
        assert _hashlen(hash) in [63, 64]

    def test_calculate_dhash_rgb_hashlen_rgba(self, sample_image_square_rgba):
        hash = calculate_dhash(sample_image_square_rgba, sqrt_hash_size=8)
        assert _hashlen(hash) in [63, 64]

    def test_calculate_dhash_rgb_hashlen_gray(self, sample_image_tall_grayscale):
        hash = calculate_dhash(sample_image_tall_grayscale, sqrt_hash_size=8)
        assert _hashlen(hash) in [63, 64]

    def test_record_dhash_hash_robust_to_resize_rgb(self, sample_image_square_rgb):
        hash = calculate_dhash(sample_image_square_rgb, sqrt_hash_size=8)
        resized_hash = calculate_dhash(
            resize(sample_image_square_rgb, shape=(24, 24)), sqrt_hash_size=8
        )
        assert _hamming_dist(hash, resized_hash) < 10

    def test_record_dhash_hash_robust_to_resize_rgba(self, sample_image_square_rgba):
        hash = calculate_dhash(sample_image_square_rgba, sqrt_hash_size=8)
        resized_hash = calculate_dhash(
            resize(sample_image_square_rgba, shape=(24, 24)), sqrt_hash_size=8
        )
        assert _hamming_dist(hash, resized_hash) < 10

    def test_record_dhash_hash_robust_to_resize_gray(self, sample_image_tall_grayscale):
        hash = calculate_dhash(sample_image_tall_grayscale, sqrt_hash_size=8)
        resized_hash = calculate_dhash(
            resize(sample_image_tall_grayscale, shape=(24, 24)), sqrt_hash_size=8
        )
        assert _hamming_dist(hash, resized_hash) < 10

    def test_record_dhash_hash_is_different_for_non_duplicates(
        self, sample_image_square_rgb, sample_image_tall_grayscale
    ):
        hash1 = calculate_dhash(sample_image_square_rgb, sqrt_hash_size=8)
        hash2 = calculate_dhash(sample_image_tall_grayscale, sqrt_hash_size=8)
        assert _hamming_dist(hash1, hash2) > 10


def _hamming_dist(hash1, hash2):
    return bin(hash1 ^ hash2).count("1")


def _hashlen(hash):
    return len(f'{hash:b}')


class TestReportDHash:
    def test_report_dhash(self, sample_image_square_rgb):
        log_dict = defaultdict(dict)
        image_path = SAMPLE_DATA_DIR / 'creevey_rgb.jpg'
        result = report_dhash(
            sample_image_square_rgb, inpath=image_path, log_dict=log_dict
        )
        assert log_dict[image_path]['dhash'] == calculate_dhash(sample_image_square_rgb)
        assert (result == sample_image_square_rgb).all()
